/*
 *		s
 *
 *		1994.5.22		T.Kobayashi
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "exec.h"
#include "err.h"

#include "inlib.h"

int			ExecDebug ;
CodeStruct	*ExecPtr ;
DataStruct	*DataGlobalVar, *DataLocalVar ;
int			ExecLine ;
char		*ExecFile ;
int			MaxStacks = 1024 ;

static	int		ExecIf( int );
static	int		ExecWhile( void );
static	int		ExecDoWhile( void );
static	int		ExecFor( void );
static	int		ExecComplex( void );
static	int		ExecBreak( void );
static	int		ExecContinue( void );

extern	int	(*BreakCheck)(void);

/*		s	*/
int		ExecSentense()
{
	int		ret = RETURN_NOMAL ;

	if ( ExecPtr == NULL )
		return RETURN_END ;

	ExecLine = ExecPtr->sent.line ;
	ExecFile = ExecPtr->sent.file ;

	if ((*BreakCheck)()) {
		ExecError("I܂");
	}

	if ( ExecDebug )
	{
		printf( "EXEC[%06X][%s:%4d] : %d : %d\n",
			ExecPtr, ExecFile, ExecLine, ExecPtr->type, ExecPtr->sent.stype );
	}

	if (ExecPtr->type != CODE_SENTENSE) {
		printf("ExecPtr = %08x\n", ExecPtr);
		fflush(stdout);
	}
	assert( ExecPtr->type == CODE_SENTENSE );

	switch( ExecPtr->sent.stype )
	{
		case SENT_DEF_FUNC:
			SkipSentense();
			break ;
		case SENT_VAR:
			ExecVar();
			break ;
		case SENT_ASSIGN:
			ExecAssign( ExecPtr->sent.stype );
			break ;
		case SENT_ASN_PLUS:
		case SENT_ASN_MINUS:
		case SENT_ASN_MULT:
		case SENT_ASN_DIVIDE:
		case SENT_ASN_AND:
		case SENT_ASN_OR:
			ExecAssignOpe( ExecPtr->sent.stype );
			break ;
		case SENT_ASN_INC:
		case SENT_ASN_DEC:
			ExecAssignInc( ExecPtr->sent.stype );
			break ;
		case SENT_IF:
			ret = ExecIf( FALSE );
			break ;
		case SENT_IF_ELSE:
			ret = ExecIf( TRUE );
			break ;
		case SENT_WHILE:
			ret = ExecWhile();
			break ;
		case SENT_DO:
			ret = ExecDoWhile();
			break ;
		case SENT_FOR:
			ret = ExecFor();
			break ;
		case SENT_EXPRESS:
			GotoNextCode();
			ExecExpression();
			StackPop();
			break ;
		case SENT_BREAK:
			ret = ExecBreak();
			break ;
		case SENT_CONTINUE:
			ret = ExecContinue();
			break ;
		case SENT_RETURN:
			ret = ExecReturn();
			break ;
		case SENT_COMPLEX:
			ret = ExecComplex();
			break ;
		case SENT_NULL:
			GotoNextCode();
			break;
		default:
			assert( FALSE );
	}
	return ret ;
}

/*
 *		XLbv
 */
void	SkipSentense()
{
	ExecPtr = CodeNextSentense( ExecPtr );
}

/*
 *		if ̎s
 */
static	int		ExecIf( sw )
int			sw ;
{
	int			ret = RETURN_NOMAL ;
	int			flag ;
	DataStruct	*buf ;

	GotoNextCode();

	/*		*/
	ExecExpression();
	buf = StackTop();
	if ( ( buf->type & (TYPE_BOOLEAN|TYPE_INT) ) == 0 )
		ExecError( "sł" );

	flag = buf->id.i ;
	StackPop();

	/*		*/
	if ( flag )
	{
		ret = ExecSentense();
		if ( sw )
			SkipSentense();
	}
	else
	{
		SkipSentense();
		if ( sw )
			ret = ExecSentense();
	}
	return ret ;
}

/*
 *		while ̎s
 */
static	int		ExecWhile()
{
	int		flag ;
	int		ret = RETURN_NOMAL ;
	CodeStruct	*condptr ;	/*	̈ʒu	*/
	DataStruct	*buf ;

	GotoNextCode();
	condptr = ExecPtr ;

	for(;;)
	{
		/*		*/
		ExecExpression();
		buf = StackTop();
		if ( ( buf->type & (TYPE_BOOLEAN|TYPE_INT) ) == 0 )
			ExecError( "sł" );
		flag = buf->id.i ;
		StackPop();

		/*		*/
		if ( flag )
		{
			ret = ExecSentense();
			if ( ret == RETURN_CONTINUE )
				ret = RETURN_NOMAL ;
			if ( ret == RETURN_BREAK )
			{
				ret = RETURN_NOMAL ;
				break ;
			}
			if ( ret == RETURN_RETURN )
				break ;
			ExecPtr = condptr ;
		}
		else
		{
			SkipSentense();
			break ;
		}
	}
	return ret ;
}

/*
 *		do-while ̎s
 */
static	int		ExecDoWhile()
{
	int		ret = RETURN_NOMAL ;
	CodeStruct	*endadr ;	/*	IAhX		*/
	CodeStruct	*sentadr ;	/*	s̃AhX	*/
	DataStruct	*buf ;

	GotoNextCode();
	endadr = (CodeStruct*)( ExecPtr->c.data.id.i );
	GotoNextCode();
	sentadr = ExecPtr ;

	for(;;)
	{
		/*	s	*/
		ret = ExecSentense();
		if ( ret == RETURN_CONTINUE )
			ret = RETURN_NOMAL ;
		if ( ret == RETURN_BREAK )
		{
			ret = RETURN_NOMAL ;
			break ;
		}
		if ( ret == RETURN_RETURN )
			break ;

		/*		*/
		ExecExpression();
		buf = StackTop();
		if ( ( buf->type & (TYPE_BOOLEAN|TYPE_INT) ) == 0 )
			ExecError( "sł" );
		StackPop();
		if ( buf->id.i )
			ExecPtr = sentadr ;
		else
			break ;
	}
	ExecPtr = endadr ;
	return ret ;
}

/*
 *		for ̎s
 */
static	int		ExecFor()
{
	int		flag ;
	int		ret = RETURN_NOMAL ;
	CodeStruct	*repptr ;	/*	JԂ̈ʒu	*/
	CodeStruct	*condptr ;	/*	̈ʒu	*/
	DataStruct	*buf ;

	GotoNextCode();

	/*		*/
	ExecSentense();
	condptr = ExecPtr ;

	for(;;)
	{
		/*		*/
		ExecPtr = condptr ;
		ExecExpression();
		buf = StackTop();
		if ( ( buf->type & (TYPE_BOOLEAN|TYPE_INT) ) == 0 )
			ExecError( "sł" );
		flag = buf->id.i ;
		StackPop();

		repptr = ExecPtr ;
		SkipSentense();

		/*		*/
		if ( flag )
		{
			ret = ExecSentense();
			if ( ret == RETURN_CONTINUE )
				ret = RETURN_NOMAL ;
			if ( ret == RETURN_BREAK )
			{
				ret = RETURN_NOMAL ;
				break ;
			}
			if ( ret == RETURN_RETURN )
				break ;
		}
		else
		{
			SkipSentense();
			break ;
		}

		/*	JԂ	*/
		ExecPtr = repptr ;
		ExecSentense();
	}
	return ret ;
}

/*
 *		̎s
 */
static	int		ExecComplex()
{
	int		ret = RETURN_NOMAL ;
	CodeStruct	*endadr ;
	CodeStruct	*code ;

	GotoNextCode();
	code = ExecPtr;
	assert( code->type == CODE_CONST );
	endadr = (CodeStruct*)( code->c.data.id.i );

	GotoNextCode();
	while( ExecPtr < endadr )
	{
		ret = ExecSentense();
		if ( ret != RETURN_NOMAL )
			break ;
	}

	if ( endadr >= CodeNextPtr() )
		ExecPtr = NULL ;
	else
		ExecPtr = endadr ;
	return ret ;
}

/*
 *		break ̎s
 */
static	int		ExecBreak()
{
	GotoNextCode();
	return RETURN_BREAK ;
}


/*
 *		continue ̎s
 */
static	int		ExecContinue()
{
	GotoNextCode();
	return RETURN_CONTINUE ;
}

